/*
 * Copyright 2014 Peter Olsson
 * 
 * This file is part of Brum Brum Rally.
 * 
 * Brum Brum Rally is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Brum Brum Rally is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "GameList.h"
#include "Network.h"
#include "StateManager.h"
#include "GameState.h"
#include <sstream>

const std::size_t MAX_VISIBLE_GAMES = 4;

const int HOVER_ARROW_UP = -2;
const int HOVER_ARROW_DOWN = -3;

GameList::GameList()
:	selected(false),
	iGame(0),
	scrollPos(0),
	gameState(0),
	hasFocus(false),
	hoverGame(-1),
	upArrowHover(false),
	upArrowVisible(false),
	downArrowHover(false),
	downArrowVisible(false),
	noGamesText("No games!", 100)
{
}

void GameList::update(const IPaddress& address, const NetworkGameInfo* gameInfo)
{
	std::size_t pos = 0;
	
	// Find position in vector.
	for (pos = 0; pos < games.size(); ++pos)
	{
		if (games[pos].address == address)
		{
			break;
		}
	}
	
	// Create if non-existing.
	if (pos >= games.size())
	{
		Game game;
		game.address = address;
		game.renderObject = new RenderObject;
		games.push_back(game);
	}
	
	Game& game = games[pos];
	
	if (gameInfo)
	{
		game.text[0] = gameInfo->name;
		game.text[1] = gameInfo->tournament ? "tournament" : "single race";
		{
			std::ostringstream oss;
			oss << "players: " << gameInfo->players << '/' << gameInfo->maxPlayers;
			game.text[2] = oss.str();
		}
		{
			std::ostringstream oss;
			oss << "ping: " << gameInfo->ping;
			game.text[3] = oss.str();
		}
		{
			std::ostringstream oss;
			oss << "collisions: " << (gameInfo->collisions ? "on" : "off");
			game.text[4] = oss.str();
		}
		
		updateRenderObject(pos);
	}
	else
	{
		if (gameState)
		{
			gameState->removeRenderObject(*game.renderObject);
		}
		delete game.renderObject;
		games.erase(games.begin() + pos);
		if (pos <= iGame)
		{
			if (scrollPos == iGame &&
			    scrollPos > 0)
			{
				--scrollPos;
			}
			--iGame;
		}
	}
	updateRenderObjects();
}

IPaddress GameList::onEvent(const SDL_Event& event)
{
	IPaddress joinAddress = IPaddress();
	
	MenuEvent me = StateManager::getInstance().getMenuEvent(event);
	switch (me.type)
	{
	case MENU_MOUSE_MOVE:
		mouseMove(me.x, me.y);
		break;
	case MENU_MOUSE_CLICK1:
		click(me.x, me.y, joinAddress);
	}
	
	switch (event.type)
	{
	case SDL_MOUSEBUTTONDOWN:
		switch (event.button.button)
		{
		case SDL_BUTTON_WHEELUP:
			scrollUp();
			break;
		case SDL_BUTTON_WHEELDOWN:
			scrollDown();
			break;
		}
	}
	
	if (hasFocus)
	{	
		switch (me.type)
		{
		case MENU_NEXT:
			next();
			break;
		case MENU_PREV:
			prev();
			break;
		case MENU_MIN:
			for (int i = iGame; i >= 0; --i)
			{
				prev();
			}
			break;
		case MENU_MAX:
			for (std::size_t i = iGame; i < games.size(); ++i)
			{
				next();
			}
			break;
		case MENU_FIRST:
			for (std::size_t i = 0; i < MAX_VISIBLE_GAMES; ++i)
			{
				if (scrollUp())
				{
					prev();
				}
			}
			break;
		case MENU_LAST:
			for (std::size_t i = 0; i < MAX_VISIBLE_GAMES; ++i)
			{
				if (scrollDown())
				{
					next();
				}
			}
			break;
		case MENU_ACTION:
			if (selected)
			{
				joinAddress = games[iGame].address;
			}
			break;
		}
	}
	return joinAddress;
}

void GameList::setGameState(GameState* newGameState)
{
	if (gameState)
	{
		for (std::size_t i = 0; i < games.size(); ++i)
		{
			gameState->removeRenderObject(*games[i].renderObject);
		}
	}
	gameState = newGameState;
	hasFocus = false;
	selected = false;
	updateRenderObject(iGame);
	updateRenderObjects();
}

void GameList::focus(bool focusValue)
{
	hasFocus = focusValue;
	selected = focusValue && !games.empty();
	updateRenderObject(iGame);
	if (games.empty())
	{
		if (hasFocus)
		{
			noGamesText.addTo(gameState);
		}
		else
		{
			noGamesText.removeFrom(gameState);
		}
	}
}

bool GameList::focus()
{
	return hasFocus;
}

void GameList::updateRenderObject(std::size_t i)
{
	if (i >= games.size())
	{
		return;
	}
	Game& game = games[i];
	
	Color color = (i == iGame && selected ? MENU_ACTIVE_COLOR : MENU_INACTIVE_COLOR);
	
	Surface s[5];
	s[0] = Surface(game.text[0], BIG_FONT, color, TEXT_BORDER_COLOR);
	
	for (i = 1; i < 5; ++i)
	{
		s[i] = Surface(game.text[i], SMALL_FONT, color, TEXT_BORDER_COLOR);
	}
	
	const int SPACING_Y = 2;
	const int INFO_W1 = 52;
	const int INFO_W2 = 60;
	int w = std::max(INFO_W1 + INFO_W2, s[0].width());
	int h = s[0].height() + 2 * (s[1].height() + SPACING_Y);
	
	Surface surface(w, h);
	surface.fill(DEFAULT_COLORKEY);
	surface.setColorKey(DEFAULT_COLORKEY);
	
	s[0].draw(surface, (w - s[0].width()) / 2, 0);
	
	int x1 = (w - (INFO_W1 + INFO_W2)) / 2;
	int x2 = x1 + INFO_W1;
	int y1 = s[0].height() + SPACING_Y;
	int y2 = y1 + s[1].height() + SPACING_Y;
	
	s[1].draw(surface, x1, y1);
	s[2].draw(surface, x2, y1);
	s[3].draw(surface, x1, y2);
	s[4].draw(surface, x2, y2);

	game.renderObject->setSurface(surface);
}

void GameList::updateRenderObjects()
{
	if (!gameState)
	{
		return;
	}
	
	if (games.empty())
	{
		if (hasFocus)
		{
			noGamesText.addTo(gameState);
		}
		return;
	}
	else if (hasFocus)
	{
		noGamesText.removeFrom(gameState);
	}
	
	
	if (iGame >= games.size())
	{
		iGame = games.size() - 1;
		updateRenderObject(iGame);
		selected = false;
	}
	
	if (iGame < scrollPos)
	{
		scrollPos = iGame;
	}
	else if (iGame >= scrollPos + MAX_VISIBLE_GAMES)
	{
		scrollPos = iGame - (MAX_VISIBLE_GAMES - 1);
	}
	
	std::size_t scrollPosEnd = scrollPos + MAX_VISIBLE_GAMES;
	if (scrollPosEnd > games.size())
	{
		scrollPos -= scrollPosEnd - games.size();
		if (scrollPos > games.size())
		{
			scrollPos = 0;
		}
	}
	
	int y = 60;
	for (std::size_t i = 0; i < games.size(); ++i)
	{
		gameState->removeRenderObject(*games[i].renderObject);
		if (i >= scrollPos && i < scrollPosEnd)
		{
			RenderObject* ro = games[i].renderObject;
			ro->moveTo((GAME_WIDTH - ro->getSurface().width()) / 2, y);
			y += ro->getSurface().height() + 5;
			gameState->addRenderObject(*ro);
		}
	}
	
	updateUpArrow(upArrowHover, scrollPos > 0);
	updateDownArrow(downArrowHover, scrollPosEnd < games.size());
}

void GameList::updateUpArrow(bool arrowHover, bool arrowVisible)
{
	if (arrowHover != upArrowHover || arrowVisible != upArrowVisible)
	{
		upArrowHover = arrowHover;
		upArrowVisible = arrowVisible;
		
		gameState->removeRenderObject(upArrow);
		if (arrowVisible)
		{
			Color color = (arrowHover ? MENU_ACTIVE_COLOR : MENU_INACTIVE_COLOR);
			upArrow.setSurface(Surface("^", BIG_FONT, color, TEXT_BORDER_COLOR).subSurface(0, 0, 7, 5));
			upArrow.moveTo((GAME_WIDTH - upArrow.getSurface().width()) / 2, 53);
			gameState->addRenderObject(upArrow);
		}
	}
}

void GameList::updateDownArrow(bool arrowHover, bool arrowVisible)
{
	if (arrowHover != downArrowHover || arrowVisible != downArrowVisible)
	{
		downArrowHover = arrowHover;
		downArrowVisible = arrowVisible;
		
		gameState->removeRenderObject(downArrow);
		if (arrowVisible)
		{
			Color color = (arrowHover ? MENU_ACTIVE_COLOR : MENU_INACTIVE_COLOR);
			downArrow.setSurface(Surface("^", BIG_FONT, color, TEXT_BORDER_COLOR).subSurface(0, 0, 7, 5));
			downArrow.moveTo((GAME_WIDTH - downArrow.getSurface().width()) / 2, 185);
			downArrow.setRotation(PI);
			gameState->addRenderObject(downArrow);
		}
	}
}

void GameList::next()
{
	if (selected)
	{
		if (iGame + 1 < games.size() )
		{
			std::size_t old = iGame;
			++iGame;
			updateRenderObject(old);
			updateRenderObject(iGame);
			updateRenderObjects();
		}
	}
	else
	{
		selected = !games.empty();
		updateRenderObject(iGame);
	}
}

void GameList::prev()
{
	if (selected)
	{
		if (iGame > 0)
		{
			std::size_t old = iGame;
			--iGame;
			updateRenderObject(old);
			updateRenderObject(iGame);
			updateRenderObjects();
		}
	}
	else
	{
		selected = !games.empty();
		updateRenderObject(iGame);
	}
}

bool GameList::scrollDown()
{
	if (scrollPos + MAX_VISIBLE_GAMES < games.size())
	{
		++scrollPos;
		if (iGame < scrollPos)
		{
			++iGame;
			selected = false;
			updateRenderObject(iGame - 1);
		}
		updateRenderObjects();
		return true;
	}
	return false;
}
bool GameList::scrollUp()
{
	if (scrollPos > 0)
	{
		--scrollPos;
		if (iGame >= scrollPos + MAX_VISIBLE_GAMES)
		{
			--iGame;
			selected = false;
			updateRenderObject(iGame + 1);
		}
		updateRenderObjects();
		return true;
	}
	return false;
}

void GameList::mouseMove(int x, int y)
{
	std::size_t scrollPosEnd = scrollPos + MAX_VISIBLE_GAMES;
	updateUpArrow(upArrow.isHit(x, y), scrollPos > 0);
	updateDownArrow(downArrow.isHit(x, y), scrollPosEnd < games.size());
	
	std::size_t scrollEnd = std::min(scrollPosEnd, games.size());
	for (std::size_t i = scrollPos; i < scrollEnd; ++i)
	{
		if (games[i].renderObject->isHit(x, y))
		{
			hoverGame = i;
			if (iGame != i || !selected || !hasFocus)
			{
				hasFocus = true;
				selected = !games.empty();
				std::size_t old = iGame;
				iGame = i;
				updateRenderObject(old);
				updateRenderObject(iGame);
			}
			return;
		}
	}
	
	// Not found.
	
	if (hoverGame != -1 && selected)
	{
		hasFocus = false;
		selected = false;
		updateRenderObject(iGame);
	}
	hoverGame = -1;
}

void GameList::click(int x, int y, IPaddress& joinAddress)
{
	if (upArrowHover)
	{
		scrollUp();
	}
	else if (downArrowHover)
	{
		scrollDown();
	}
	else 
	{
		std::size_t scrollEnd = std::min(scrollPos + MAX_VISIBLE_GAMES, games.size());
		for (std::size_t i = scrollPos; i < scrollEnd; ++i)
		{
			if (games[i].renderObject->isHit(x, y))
			{
				if (iGame != i || !selected || !hasFocus)
				{
					hasFocus = true;
					selected = !games.empty();
					std::size_t old = iGame;
					iGame = i;
					updateRenderObject(old);
					updateRenderObject(iGame);
				}
				else
				{
					joinAddress = games[i].address;
				}
				return;
			}
		}
		hasFocus = false;
	}
}
